"""
Singleton class that contains:
    objectives dict
    tags dict
    teams dict
    functions dict

    constobj
    prefix
"""

import logging
import os.path
from collections import deque
from types import MappingProxyType

if __name__ == "__main__":
    import sys
    sys.path.append("..")
    del sys

from fenalib.assert_utils import assert_type
from fenalib.repr_utils import addrepr
from fenalib.mcfunction import McFunction
from fenalib.config_data import ConfigData, get_all_data

config_data = ConfigData()

def get_mcfunc_directories(full_path):
    """
    Args:
        full_path (str): The full path of the mcfunction

    Returns:
        deque: All folders used to specify the path of the mcfunction in minecraft
    """
    # strips away ".mcfunction"
    path_without_ext, extension = os.path.splitext(full_path)
    assert extension == ".mcfunction"

    # gets the list of all directories including base file without extension
    path_list = os.path.normpath(path_without_ext).split(os.sep)

    # gets all directories of the shortcut including the function name
    # path_list should contain "functions" by the end, or else it will be an empty list
    directories = deque()
    while path_list and path_list[-1] != "functions":
        directories.appendleft(path_list.pop())

    if not path_list:
        raise SyntaxError("Path {} must contain a functions/ folder".format(full_path))

    if len(directories) <= 1 and config_data.version == "1.12":
        raise SyntaxError("Path {} must have a folder inside the functions/ folder".format(full_path))

    if config_data.version != "1.12":
        # gets the folder after the "functions" folder as the main namespace"
        path_list.pop()
        directories.appendleft(path_list.pop())

    return directories


def _get_all_shortcuts(directories):
    """
    Args:
        directories (deque of str objects): All directories up to but excluding the function folder

    Returns:
        list: All possible shortcuts to the mcfunction path
        str: The mcfunction path as specified by minecraft

    Examples:
        >>> i = InFileConfig()
        >>> directories = deque(["ego", "floo_network", "init"])
        >>> i._get_all_shortcuts(directories)
        (['ego:floo_network/init', 'floo_network/init', 'init'], 'ego:floo_network/init')
    """
    # gets the mcfunction path
    mcfunction_path = directories.popleft() + ":"
    mcfunction_path += "/".join(directories)

    # shortcuts also has the mcfunction path to map to itself to pass the FunctionBuilder containment test
    shortcuts = []
    shortcuts.append(mcfunction_path)

    # gets all shortcuts to the full name
    while directories:
        shortcut = "/".join(directories)
        shortcuts.append(shortcut)
        directories.popleft()

    return shortcuts, mcfunction_path

@addrepr
class InFileConfig:
    """
    Holds branching simple commands using a builder model
    Args:
        prefix (str): Any prefix to objectives, teams or tags
        constobj (str): The objective used for scoreboard players operation with a constant value
        functions (dict): All function shortcuts present in the file
            - Maps any version of the function shortcut to the function path
        function_conflicts (set): pass
    """
    default_prefix = "fena"
    default_constobj = "g.number"

    def __init__(self):
        self._functions = {}
        self._function_conflicts = set()
        self._prefix = None
        self._constobj = None
        self.finalized = False

    def __new__(cls):
        """
        Ensures they are the same class
        """
        if not hasattr(cls, '_in_file_config_data'):
            cls._in_file_config_data = super().__new__(cls)
        return cls._in_file_config_data

    @property
    def prefix(self):
        return self._prefix

    @prefix.setter
    def prefix(self, prefix):
        """
        Does not allow the constobj to be set multiple times

        Args:
            prefix (str)
        """
        if self._prefix is not None:
            raise SyntaxError("{}: Cannot set a prefix twice".format(prefix))

        assert_type(prefix, str)
        self._prefix = prefix

    @property
    def constobj(self):
        return self._constobj

    @constobj.setter
    def constobj(self, constobj):
        """
        Does not allow the constobj to be set multiple times

        Args:
            prefix (str)
        """
        if self._constobj is not None:
            raise SyntaxError("{}: Cannot set a constobj twice".format(constobj))

        assert_type(constobj, str)
        self._constobj = constobj

    @property
    def functions(self):
        return self._functions

    @property
    def function_conflicts(self):
        return self._function_conflicts

    def add_function(self, mcfunction):
        """
        Gets the function paths all the way until "functions"
            - If the version is greater than 1.13, this also gets the folder before "functions" as the namespace

        Args:
            mcfunction (McFunction)
        """
        assert_type(mcfunction, McFunction)
        full_path = mcfunction.full_path
        directories = get_mcfunc_directories(full_path)
        shortcuts, mcfunction_path = _get_all_shortcuts(directories)

        # updates the functions dictionary to map the entire list of shortcuts to the mcfunction_path string
        for shortcut in shortcuts:
            if shortcut in self._functions:
                self._functions.pop(shortcut, None)
                self._function_conflicts.add(shortcut)
            else:
                self._functions[shortcut] = mcfunction_path

    def finalize(self):
        """
        Sets the prefix and the constobj to their default values if none were found
        """
        assert not self.finalized

        # default for prefix is "fena"
        if self._prefix is None:
            self._prefix = InFileConfig.default_prefix
            logging.warning(f"Using the default prefix {self._prefix!r}")

        # default for constobj is "g.number"
        if self._constobj is None:
            self._constobj = InFileConfig.default_constobj
            if not config_data.ego:
                logging.warning(f"Using the default constobj {self._constobj!r}")

        # sets previously mutable types as immutable
        self._functions = MappingProxyType(self._functions)
        self._function_conflicts = frozenset(self._function_conflicts)

        self.finalized = True


if __name__ == "__main__":
    # mcfunction = McFunction(r"C:\Users\Austin-zs\Documents\Austin\powder game code\Programming\CCU\functions\ego\floo_network\init.mcfunction")
    # in_file_config = InFileConfig()

    # in_file_config.add_function(mcfunction)
    # print(in_file_config.functions)
    # import doctest
    # doctest.testmod()

    command_template = (
        "say test1 inside function {mfunc_name}",
        "say test2 inside function {mfunc_name}",
        "say test3 inside function {mfunc_name}",
    )

    def add_function(commands, function_name):
        return tuple(f"{command} with function {function_name}" for command in commands)

    def get_full_path(path):
        dir_path = os.path.dirname(os.path.realpath(__file__))
        return os.path.join(dir_path, path) + ".mcfunction"

    def get_mfunc_name(full_path):
        # wtf is this code even
        # substring to the left of the last instance of functions
        # removes the first character of the string
        # replaces the first instance of '/' with ':'
        return full_path.split("functions")[-1][1:].replace("/", ":", 1)

    def create_mfunc(relative_path):
        full_path = get_full_path(relative_path)
        mfunc_name = get_mfunc_name(full_path)

        mcfunction = McFunction(mfunc_name, full_path)
        for command in command_template:
            mcfunction.add_command(command.format(mfunc_name=mfunc_name))
        mcfunction.finalize()

        return mcfunction

    # get_all_data(version="1.12")
    # mcfunc1 = create_mfunc("functions/ego/test/test1")
    # mcfunc2 = create_mfunc("functions/ego/test/test2")
    # mcfuncwithin1 = create_mfunc("functions/ego/test/within_test/test1")
    # mcfuncwithin2 = create_mfunc("functions/ego/test/within_test/test2")
    # mcfuncwithin22 = create_mfunc("functions/ego/test/within_test2/test2")

    get_all_data(version="1.13")
    mcfunc1 = create_mfunc("ego/functions/test/test1")
    mcfunc2 = create_mfunc("ego/functions/test/test2")
    mcfuncwithin1 = create_mfunc("ego/functions/test/within_test/test1")
    mcfuncwithin2 = create_mfunc("ego/functions/test/within_test/test2")
    mcfuncwithin22 = create_mfunc("ego/functions/test/within_test2/test2")

    in_file_config = InFileConfig()
    in_file_config.add_function(mcfunc1)
    in_file_config.add_function(mcfunc2)
    in_file_config.add_function(mcfuncwithin1)
    in_file_config.add_function(mcfuncwithin2)
    in_file_config.add_function(mcfuncwithin22)
    print(repr(in_file_config))





import re
import os
from typing import NamedTuple, Optional

if __name__ == "__main__":
    import sys
    sys.path.append("..")
    del sys

from fenalib.assert_utils import assert_type
from fenalib.str_utils import decode_str

r"""
Runs a quick script before pyexpander is run
Features:
    - replaces all \\ with \\\\
    - changes some syntax of pyexpander code to make it more elegant to work with for fena code
    - removes all empty lines and comments (although a trailing newline is added)

Syntax changes:

```
$py(
code = "unchanged"
)

$py:
    idk = []
    for x in range(15):
        idk.append(x*2)

!mfunc hello
    $py:
        somedict = {}
        for x in range(15):
            somedict[x] = "ayylmao"

$macro(edit_give_porkchops):
    say this is a macro

$macro(unchanged)
    say this is an unchanged macro
$endmacro
------ TO ------
$py(
code = "unchanged"
)

$py(
idk = []
for x in range(15):
    idk.append(x*2)
)

!mfunc hello
$py(
somedict = {}
for x in range(15):
    somedict[x] = "ayylmao"
)

$macro(edit_give_porkchops)
say this is a macro
$endmacro

$macro(unchanged)
    say this is an unchanged macro
$endmacro
```

# for loops and anything else with $keyword() and $endkeyword
$for(x in range(5)):
    !mfunc give_porkchop_$(x)
        give @s minecraft:porkchop $(x)

!mfunc give_porkchops
    $for(x in range(5)):
        @s[porkchop=$(x)] function give_porkchop_$(x)

    $for(_ in range(3))
    $("say unchanged code")
    $endfor

------ TO ------
$for(x in range(5))
!mfunc give_porkchop_$(x)
    give @s minecraft:porkchop $(x)
$endfor

!mfunc give_porkchops
    $for(x in range(5))
    @s[porkchop=$(x)] function give_porkchop_$(x)
    $endfor

    $for(_ in range(3))
    $("say unchanged code")
    $endfor
```
"""

KEYWORD_SURROUND = {
    "if": "$endif",
    "for": "$endfor",
    "while": "$endwhile",
    "for_begin": "$endfor",
    "while_begin": "$endwhile",
    "macro": "$endmacro",
}

LINESEP = "\n"

INDENT = "    "

class PrePyexpanderError(Exception):
    pass

class InvalidFileError(PrePyexpanderError):
    pass

class InvalidIndentError(PrePyexpanderError):
    pass

class Line(NamedTuple):
    """
    Attributes:
        string (str): A fully unindented string as the line
        indents (int): The number of indents that the string had
        line_number (int): The line number in the original file
        pyexp_keyword (Optional[str]): The pyexpander keyword if it has the proper syntax for pre-expander parsing
            - line starts with '$' and ends with ':'
        empty (bool): Whether the line should be considered for parsing indents or not
    """
    string: str
    line_number: int
    indents: int
    pyexp_keyword: Optional[str]
    empty: bool


def get_line_strings(text):
    """
    Generates all lines from a text (without the newline)
    """
    begin_index = 0
    for index, char in enumerate(text):
        if char == LINESEP:
            yield text[begin_index:index]
            begin_index = index + 1

rx_INDENT_WHITESPACE = re.compile(r'^\s*')  # gets left most whitespace
rx_PYEXP_KEYWORD = re.compile(r'[a-zA-Z]+')  # pyexpander keyword
def get_lines(text):
    """
    Generates Line objects from the text
    """
    for line_number, line_str in enumerate(get_line_strings(text)):

        # if it's a comment or empty line, the line is considered to be empty
        empty = (line_str.lstrip().startswith("#") or not line_str.strip())

        if empty:
            yield Line(line_str, line_number, indents=0, pyexp_keyword=None, empty=True)
            continue

        # checks for proper indentation as 4 spaces
        match = re.match(rx_INDENT_WHITESPACE, line_str)
        indent_str = "" if match is None else match.group()
        if "\t" in indent_str:
            raise InvalidIndentError(f"Line {line_number} with indentation of {indent_str!r} contains a tab space in line {line_str!r}")
        if indent_str.count(" ") % len(INDENT) != 0:
            raise InvalidIndentError(f"Line {line_number} with indentation of {indent_str!r} requires indentation with 4 spaces in line {line_str!r}")

        indents = indent_str.count(INDENT)
        unindented_str = line_str.lstrip()

        # checks if it contains a pyexpander keyword that should be parsed by indentation
        if unindented_str.startswith("$") and unindented_str.endswith(":"):
            # gets python keyword
            keyword = rx_PYEXP_KEYWORD.match(unindented_str, 1)
            if keyword is None:
                raise PrePyexpanderError(f"Line {line_number}: Expected a pyexpander keyword in line {line_str!r}")
            yield Line(unindented_str, line_number, indents, pyexp_keyword=keyword.group(), empty=False)
        elif unindented_str.startswith("$include(") and unindented_str.endswith(")"):
            file_path = unindented_str[len("$include("):-len(")")]
            yield Line(file_path, line_number, indents, pyexp_keyword="include", empty=False)
        else:
            yield Line(unindented_str, line_number, indents, pyexp_keyword=None, empty=False)


def finalize_line(line, dedents=0):
    """
    Indents the line given line indents minus the dedents, and removes the ending ":" if it exists
    """
    assert_type(line, Line)
    assert_type(dedents, int)
    total_indents = line.indents - dedents

    if total_indents < 0:
        raise InvalidIndentError(f"Invalid indentation for line {line} with dedents={dedents}")

    if line.pyexp_keyword is not None:
        line_str = line.string[:-1]
    else:
        line_str = line.string

    return indent_line(line_str, total_indents)

def indent_line(line_str, indents):
    """
    Indents the line given indents
    """
    assert_type(line_str, str)
    assert_type(indents, int)
    return (INDENT * indents) + line_str


class Parser:
    """
    Attributes:
        reached_eof (bool): True when the final line has been reached
        current_line (str or None): The current line from the iterator (None at the end)
        iterator (generator)
    """
    def __init__(self, text):
        self.reached_eof = False
        self.current_line = None
        self.iterator = get_lines(text)
        self.advance()

    def advance(self):
        previous_line = self.current_line
        self.current_line = next(self.iterator, None)

        if self.current_line is None:
            if self.reached_eof:
                raise PrePyexpanderError("Cannot advance past the last line")
            self.reached_eof = True

        return previous_line

    def parse_lines(self, required_indents=0, dedents=0):
        """
        Generates the properly parsed lines from the text
        Note that this is recursive

        Base case is when the line indent is less than the given indent
        """
        while not self.reached_eof:
            if self.current_line.empty:
                # yield self.advance().string
                self.advance()
                continue

            if self.current_line.indents < required_indents:
                # ends once the line has less indents
                return

            elif self.current_line.pyexp_keyword == "py":
                original_indents = self.current_line.indents
                yield finalize_line(self.advance(), dedents=original_indents) + "("
                yield from self.parse_lines(original_indents+1, original_indents+1)
                yield ")"

            elif self.current_line.pyexp_keyword == "macro":
                original_indents = self.current_line.indents
                yield finalize_line(self.advance(), dedents=original_indents)
                yield from self.parse_lines(original_indents+1, original_indents+1)
                yield KEYWORD_SURROUND["macro"]

            elif self.current_line.pyexp_keyword in ("for", "for_begin", "while", "while_begin"):
                original_indents = self.current_line.indents
                keyword = self.current_line.pyexp_keyword # redefine since 'self.current_line' will change
                yield finalize_line(self.advance(), dedents=dedents)
                yield from self.parse_lines(original_indents+1, dedents+1)
                yield indent_line(KEYWORD_SURROUND[keyword], original_indents-dedents)

            elif self.current_line.pyexp_keyword == "if":
                original_indents = self.current_line.indents
                yield finalize_line(self.advance(), dedents=dedents)
                yield from self.parse_lines(original_indents+1, dedents+1)

                while self.current_line.pyexp_keyword == "elif":
                    yield finalize_line(self.advance(), dedents=dedents)
                    yield from self.parse_lines(original_indents+1, dedents+1)

                if self.current_line.pyexp_keyword == "else":
                    yield finalize_line(self.advance(), dedents=dedents)
                    yield from self.parse_lines(original_indents+1, dedents+1)

                yield indent_line(KEYWORD_SURROUND["if"], original_indents-dedents)

            elif self.current_line.pyexp_keyword == "elif":
                raise PrePyexpanderError(f"Unexpected 'elif' (without accompanying 'if') in self.current_line {self.current_line}")

            elif self.current_line.pyexp_keyword == "else":
                raise PrePyexpanderError(f"Unexpected 'else' (without accompanying 'if') in self.current_line {self.current_line}")

            elif self.current_line.pyexp_keyword == "include":
                # override's pyexpander's "include" keyword
                file_path = decode_str(self.current_line.string)
                if not os.path.isfile(file_path):
                    raise InvalidFileError(f"{file_path!r} is not a valid file path")
                with open(file_path) as file:
                    parser = Parser(file.read())
                    yield from parser.parse_lines()
                    self.advance()

            elif self.current_line.pyexp_keyword is not None:
                raise PrePyexpanderError(f"Unexpected pyexpander keyword in self.current_line {self.current_line}")

            else:
                yield finalize_line(self.advance(), dedents=dedents)


def parse_pre_pyexpander(text):
    """
    Returns:
        Properly parsed text to be used by pyexpanxer
    """
    parser = Parser(text)
    # return list(parser.parse_lines())
    text = "\n".join(parser.parse_lines()) + "\n"

    # 2 // -> 4 //// for entire text
    # replaces after the parsing to make sure the $include() files are included
    text = text.replace('\\\\', '\\\\\\\\')
    return text




"""
A butt load of code shamelessly taken from Ruslan Spivak's tutorial for making an interpreter:
    https://ruslanspivak.com/lsbasi-part1/

Pyexpander code shamelessly taken from:
    http://pyexpander.sourceforge.net/

Lots of config files shamelessly gotten from:
    https://github.com/PepijnMC/Minecraft
    https://github.com/Arcensoth/mcdata
"""

"""
Random stuff only for me honestly (vim session tab numbers):
    1: Python3/fena
        fena.py
    2: Fena/lib
        floo_event.py
        location.py
    3: floo_network/src
        main.py
        main.fena
    4: Python3/fena/fenalib/log
        after_pre_pyexpander.txt
        after_pyexpander.txt
        parsed_cmds.txt
    5: Fena.wiki
        Code-Style.md
        Code-Documentation.md
"""

import argparse
import logging
import sys
import os

# import pyexpander.lib as pyexpander
import fena_pyexpander.lib as pyexpander

import fenalib.logging_setup as logging_setup
logging_setup.setup_logging()

from fenalib.config_data import get_all_data
from fenalib.pre_pyexpander import parse_pre_pyexpander
from fenalib.lexer import Lexer
from fenalib.parser import Parser
from fenalib.interpreter import Interpreter
from fenalib.writer import Writer, write_after_pre_pyexpander, write_after_pyexpander

semantic_version = "6.0.0"
public_version = "0.3.0-ALPHA"

def get_args():
    # Usage: main.py fileName [output_path] [-d, --debug] [-c, --clean]
    parser = argparse.ArgumentParser()

    # requires input file
    parser.add_argument("file_name", help="main fena file")

    # optional output directory
    default_output_path = "functions/fena"
    parser.add_argument("output_path", nargs="?", default=default_output_path,
                        help="the directory in which all mcfunctions will be created, with the default being 'functions/fena'")

    # cleans mcfunctions
    parser.add_argument("-c", "--clean", help="removes all mcfunction files inside the output directory", action="store_true")

    # debug
    parser.add_argument("-d", "--debug", help="puts say commands at the front of each mcfunction to show who is running it", action="store_true")

    # debug log
    # parser.add_argument("-dl", "--debug-log", help="outputs the debug log to see all debug info from the Fena language", action="store_true")

    # custom version
    parser.add_argument("-v", "--version", nargs="?", default=None, help="custom version to override the config file if provided")

    args = parser.parse_args()
    args.output_path = os.path.realpath(args.output_path)
    return args


def parse_text(text, file_name, output_path, clean=False, debug=False, use_pre_pyexpander=True, write_functions=True):
    """
    Args:
        file_name (str): File name of the file that will be parsed
            - used for getting the full path to the file and setting the filename for pyexander
        output_path (str): Output path of the mcfunctions
        clean (bool): Whether mcfunctions will be deleted or not
        debug (bool): Whether mcfunctions will be debugged with a say cmd or not
        use_pre_pyexpander (bool): Whether the pre-pyexpander script will be used or not
        write_functions (bool): Whether mcfunctions should be written or not

    Returns:
        list of McFunction objects
    """

    if use_pre_pyexpander:
        text = parse_pre_pyexpander(text)
        write_after_pre_pyexpander(text)

    # add the directory of the file to the import path
    # allows any py commands in the file to import py files relative to said file
    dir_path = os.path.dirname(os.path.realpath(file_name))
    sys.path.append(dir_path)

    # raises an exception if pyexpander fails to parse
    text = pyexpander.expandToStr(text, filename=file_name, auto_indent=True, auto_indent_python=True)[0]
    write_after_pyexpander(text)

    lexer = Lexer(text)
    parser = Parser(lexer)
    interpreter = Interpreter(parser, output_path)
    mcfunctions = interpreter.interpret()

    if write_functions:
        writer = Writer(mcfunctions, clean, debug)
        writer.write()

    return mcfunctions


def main():
    print(f"Fena: semantic_version={semantic_version}, public_version={public_version}")

    # checks for python version (must be greater than or equal to 3.6)
    if not (sys.version_info.major >= 3 and sys.version_info.minor >= 6):
        raise RuntimeError("Python version must be 3.6")

    # gets all command line arguments
    args = get_args()

    #  overrides the version if necessary
    if args.version is not None:
        get_all_data(args.version)

    file_name = args.file_name
    output_path = args.output_path
    with open(file_name) as file:
        text = file.read()

    # sets the file name for logging
    logging_setup.format_file_name(file_name)
    parse_text(text, file_name, output_path, clean=args.clean, debug=args.debug)


if __name__ == '__main__':
    # print("test")

    try:
        main()
    # pylint: disable=broad-except
    except Exception as e:
        logging.exception(e) # type: ignore




import logging

if __name__ == "__main__":
    import sys
    sys.path.append("..")
    del sys

    import fenalib.logging_setup as logging_setup
    logging_setup.setup_logging()

from fenalib.assert_utils import assert_type
from fenalib.repr_utils import addrepr
from fenalib.token_classes import TypedToken, DelimiterToken, WhitespaceToken, TokenValues
from fenalib.config_data import ConfigData
from fenalib.lexical_token import Token
from fenalib.token_position import TokenPosition, TokenPositionRecorder

@addrepr
class Lexer:
    """
    Args:
        text (str)

    Attributes:
        text (str)
        recorder (TokenPositionRecorder)
        reached_eof (bool)
        indents (int)
    """
    config_data = ConfigData()

    def __init__(self, text):
        self.text = text
        self.recorder = TokenPositionRecorder()
        self.reached_eof = False

        # list of indent strings
        # 1 indent is either 4 spaces or 1 tab space
        self.indents = 0

        # bool returns true if it is a zero length string or complete whitespace
        if not self.text.strip():
            raise EOFError("File does not contain anything")

    def __iter__(self):
        """
        Breaks apart the entire document into a sequence of tokens.

        Yields:
            Token: For each token that exists in the document
        """
        yield from self.handle_line_begin()

        # does not require "else" since if it is a selector, it should end in the block above
        while not self.reached_eof:
            # skips all whitespace until \n
            if self.get_char().isspace() and not self.current_chars_are(WhitespaceToken.NEWLINE.value):
                self.skip_whitespace()
                continue

            # handles indents and dedents after newline
            elif self.current_chars_are(WhitespaceToken.NEWLINE.value):
                yield from self.handle_newline()

            else:
                yield from self.get_command()

        # at the end of file by this point, and it must end with dedent tokens if a newline did not end
        while self.indents > 0:
            yield self.get_dedent()

        # gets one last EOF token
        yield self.create_new_token(WhitespaceToken.EOF)

    def advance(self, increment=1):
        """
        Advances the position

        Args:
            increment (int): Number of chars that will be incremented
        """
        assert_type(increment, int)

        # while loop to increment the self.recorder variable
        while increment > 0 and not self.reached_eof:

            # if the current character is \n, goes to a new line
            # note that the position increments after this, meaning
            # that "\n" is actually the previous character
            if self.current_chars_are(WhitespaceToken.NEWLINE.value):
                self.recorder.increment_row()
            else:
                self.recorder.increment_column()

            if self.recorder.char_pos > len(self.text) - 1:
                self.reached_eof = True

            increment -= 1

    def error(self, message=None, expected=None):
        """
        Args:
            class_name (str): Name of the class that will implement this
            message (str or None): Message to be displayed with the error.
                Defaults to None, which will display a generic message containing the current character
        """
        # message and expected cannot both be used
        if message is None:
            char = self.get_char()
            if expected is None:
                message = f"Invalid character {char!r}"
            else:
                message = f"Expected {expected!r} but got {char!r}"
        raise TypeError(f"Lexer{self.recorder}: {message}")
        # raise TypeError(f"{self!r}: {message}")

    def invalid_method(self, *args, **kwargs):
        raise NotImplementedError(f"Lexer{self.recorder}: Invalid method")

    def skip_whitespace(self, skip_newline=False):
        """
        Skips all whitespace that isn't a newline
        """
        if skip_newline:
            while not self.reached_eof and self.get_char().isspace():
                self.advance()
        else:
            while not self.reached_eof and self.get_char().isspace() and not self.current_chars_are(WhitespaceToken.NEWLINE.value):
                self.advance()

    def advance_chars(self, chars):
        """
        Advances given the number of characters

        Args:
            chars (str)
        """
        assert_type(chars, str)
        self.advance(len(chars))

    def get_char(self):
        """
        Args:
            length (int, optional) number of characters from the current position

        Returns:
            int: current characters from the current position given the length
        """
        if self.reached_eof:
            self.error("Cannot get a character when the end of file has been reached")

        char_pos = self.recorder.char_pos
        return self.text[char_pos]

    def get_chars(self, length):
        """
        Args:
            length (int) number of characters from the current position

        Returns:
            str: current characters from the current position given the length
        """
        if self.reached_eof:
            self.error("Cannot get a character when the end of file has been reached")

        assert_type(length, int)
        char_pos = self.recorder.char_pos
        return self.text[char_pos: char_pos + length]

    def get_locked_chars(self):
        """
        Returns:
            str: current characters from the locked position to the current position
        """
        begin, end = self.recorder.locked_char_pos
        return self.text[begin:end]

    def current_chars_are(self, chars):
        """
        Looks ahead to compare characters

        Args:
            chars (str): characters provided to compare to the current string

        Returns
            bool: whether the characters provided equal to the current string
        """
        assert_type(chars, str)
        length = len(chars)
        return chars == self.get_chars(length)

    def create_new_token(self, token_type, value=None, position=None, advance=False, unlock=False):
        """
        Creates a new token and automatically fills out the token fields unless specified

        If the current state of the position is locked, it gets the locked chars instead

        Args:
            token_type (any TokenType)
            value (any, defaults to None): What value the token should have
            position (TokenPosition, defaults to None): The position of the current token
            advance (bool, defaults to False): Whether to advance after creating the token or not
            unlock (bool, defaults to False): Whethre to unlock after creating the token or not
        """
        if position is None:
            position = self.recorder.position
        assert_type(position, TokenPosition)

        if self.recorder.locked:
            value = self.get_locked_chars()

        token = Token(position, token_type, value)

        if unlock:
            self.recorder.unlock()

        if advance:
            self.advance_chars(token.value)

        return token

    def skip_comment(self):
        while not self.reached_eof and not self.current_chars_are(WhitespaceToken.NEWLINE.value):
            self.advance()

    def get_indent(self):
        """
        Gets a singular indent token and adds one from the indent length

        Returns:
            Token: Indent token
        """
        self.indents += 1
        return self.create_new_token(WhitespaceToken.INDENT)

    def get_dedent(self):
        """
        Gets a singular dedent token and subtracts one from the indent length

        Returns:
            Token: Dedent token
        """
        self.indents -= 1
        return self.create_new_token(WhitespaceToken.DEDENT)

    def handle_newline(self):
        """
        Handles the combination of a newline and the beginning of the next line

        Yields:
            Token: Newline token and all tokens gotten from handling a beginning of a line
        """
        # gets one newline token whenever called unless it is at the beginning of the line
        yield self.create_new_token(WhitespaceToken.NEWLINE)
        self.advance()

        yield from self.handle_line_begin()

    def handle_line_begin(self):
        """
        Handles all tokens gotten at the beginning of a line

        It creates the given number of indent or dedent tokens depending on the indents found.

        Yields:
            Token: Indent and dedent tokens, and statement tokens
        """
        # gets any whitespace past the newline
        self.recorder.lock()
        self.skip_whitespace()
        whitespace = self.get_locked_chars()
        self.recorder.unlock()

        if self.reached_eof:
            return

        # if the current char is a comment after whitespace, it is still an empty line
        if self.get_char() in (WhitespaceToken.NEWLINE.value, WhitespaceToken.COMMENT.value):
            if self.current_chars_are(WhitespaceToken.COMMENT.value):
                self.skip_comment()
            return

        # otherwise, the whitespace is valid for interpretation for indents and dedents
        # checks whether the indenting whitespace is actually valid (4 spaces)
        # boolean value of any integer is False if 0, True for anything else
        if len(whitespace) % len(WhitespaceToken.INDENT.value):
            self.error(f"Invalid whitespace: {whitespace!r} (requires indentation of 4 spaces)")

        # gets the number of indents at the current area
        current_indents = whitespace.count(WhitespaceToken.INDENT.value)

        if current_indents > self.indents:
            if current_indents-1 == self.indents:
                yield self.get_indent()

            else:
                indent_num = current_indents - self.indents
                self.error(f"Too many indents (Expected one, got {indent_num}")

        # gets all possible dedent tokens
        while current_indents < self.indents:
            yield self.get_dedent()

        # all possible dedent and indent tokens have been gotten by this point
        # however, there are certain things that can only happen after a valid newline (statements)
        if self.current_chars_are(DelimiterToken.EXCLAMATION_MARK.value):
            yield from self.get_statement()

    def get_statement(self):
        """
        Gets all proper statement tokens, starting from "!" and ending at the next "\n"
            This expects the first value to be "!", and the second value to be a statement keyword

        Yields:
            Token: All statement tokens (statement specifier, statement keyword, strings after keywords)
        """
        yield self.create_new_token(DelimiterToken.EXCLAMATION_MARK, advance=True)
        self.skip_whitespace()

        # assumes the rest are strings until newline
        while not self.reached_eof and self.get_char() != "\n":
            if self.get_char().isspace():
                self.skip_whitespace()
                continue

            if self.current_chars_are(DelimiterToken.QUOTE.value):
                yield self.get_literal_string()
            elif self.get_char() in TokenValues.get(DelimiterToken):
                yield self.create_new_token(DelimiterToken(self.get_char()), advance=True)
            else:
                yield self.get_until_space(exempt_chars=TokenValues.get(DelimiterToken))

    def get_until_space(self, exempt_chars=""):
        """
        Gets the string and stops when it is a whitespace character or the current
        character is within the exempt chars
        """
        if self.get_char() in exempt_chars:
            self.error(f"Cannot get a string because the string starts with {self.get_char()!r}, but cannot be within {exempt_chars}")
        if self.get_char().isspace():
            self.error(f"Cannot get a string because the string starts a whitespace {self.get_char()!r}")
        if self.reached_eof:
            self.error("Cannot get a string because the end of the file has been reached")

        self.recorder.lock()
        while not self.reached_eof and not self.get_char().isspace() and not self.get_char() in exempt_chars:
            self.advance()

        # checks if the result is a 0 length string: error
        assert self.get_locked_chars(), "Cannot be a 0 length string bruh wtf is this"
        return self.create_new_token(TypedToken.STRING, unlock=True)

    def get_command(self):
        """
        Gets all command related tokens

        Yields:
            Token: Delimiter, NBT, Selector, String tokens
        """
        while not self.reached_eof and not self.current_chars_are(WhitespaceToken.NEWLINE.value):
            if self.get_char().isspace():
                self.skip_whitespace()
                continue

            if self.current_chars_are(DelimiterToken.AT.value):
                yield from self.get_selector()

            elif self.current_chars_are(DelimiterToken.OPEN_CURLY_BRACKET.value):
                yield from self.get_curly_bracket_tag()

            elif self.current_chars_are(DelimiterToken.OPEN_SQUARE_BRACKET.value):
                yield from self.get_block_states()

            # elif self.current_chars_are(DelimiterToken.QUOTE.value):
            #     yield self.get_literal_string()

            elif self.get_chars(2) in TokenValues.get(DelimiterToken):
                yield self.create_new_token(DelimiterToken(self.get_chars(2)), advance=True)
            elif self.get_char() in TokenValues.get(DelimiterToken) - {"="}:
                yield self.create_new_token(DelimiterToken(self.get_char()), advance=True)

            else:
                yield self.get_until_space(exempt_chars=TokenValues.get(DelimiterToken) - {"="})

    def get_selector(self):
        """
        Gets all selector tokens (handles selector variables and brackets)

        Yields:
            Token: Selector tokens (Selector variables and all selector arguments)
        """
        # checks for valid selector var as a, e, r, s, p (or whatever is specified in the config)
        yield self.create_new_token(DelimiterToken.AT, advance=True)

        if self.get_char() not in Lexer.config_data.selector_variable_specifiers:
            self.error(f"Invalid selector variable: '@{self.get_char()}'")

        yield self.create_new_token(TypedToken.SELECTOR_VARIABLE_SPECIFIER, value=self.get_char(), advance=True)

        # starts getting the tokens inside square brackets including the square brackets
        if not self.reached_eof and self.current_chars_are(DelimiterToken.OPEN_SQUARE_BRACKET.value):
            yield self.create_new_token(DelimiterToken.OPEN_SQUARE_BRACKET, advance=True)
            yield from self.get_selector_arguments()
            yield self.create_new_token(DelimiterToken.CLOSE_SQUARE_BRACKET, advance=True)

    def get_selector_arguments(self):
        """
        Gets all selector arguments inside the square brackets []
        """
        while not self.current_chars_are(DelimiterToken.CLOSE_SQUARE_BRACKET.value):
            if self.get_char().isspace() and self.get_char() != WhitespaceToken.NEWLINE.value:
                self.skip_whitespace()
                continue

            elif self.current_chars_are(DelimiterToken.OPEN_CURLY_BRACKET.value):
                yield from self.get_curly_bracket_tag()

            elif self.current_chars_are('"'):
                yield self.get_literal_string()
            elif self.get_char().isdigit() or self.get_char() == "-":
                yield self.get_number()

            # Gets simple delimiter tokens
            elif self.get_char() in TokenValues.get(DelimiterToken):
                yield self.create_new_token(DelimiterToken(self.get_char()), advance=True)
            elif self.get_chars(2) in TokenValues.get(DelimiterToken):
                yield self.create_new_token(DelimiterToken(self.get_chars(2)), advance=True)

            elif self.current_chars_are("*"):
                yield self.create_new_token(TypedToken.STRING, value=self.get_char(), advance=True)

            # otherwise, general string
            elif self.get_char().isalpha() or self.get_char() in "_.":
                yield self.get_until_space(exempt_chars=TokenValues.get(DelimiterToken))

            else:
                self.error()

    def get_literal_string(self, allow_newline=False):
        """
        Returns:
            Token: Literal string with quotes enclosing it
        """
        self.recorder.lock()

        # advances the first "
        if not self.current_chars_are(DelimiterToken.QUOTE.value):
            self.error(expected=DelimiterToken.QUOTE.value)
        self.advance()

        while not self.current_chars_are(DelimiterToken.QUOTE.value):
            # Skips past any \" or \\ since it doesn't count as a string end
            if self.current_chars_are(r'\"') or self.current_chars_are(r'\\'):
                self.advance(2)

            # There shouldn't ever be a newline in a string
            elif self.current_chars_are(WhitespaceToken.NEWLINE.value) and not allow_newline:
                self.error("Unexpected newline during a string declaration")

            # Advances normally
            else:
                self.advance()

        # advances the last "
        self.advance()

        return self.create_new_token(TypedToken.LITERAL_STRING, unlock=True)

    def get_number(self):
        """
        Gets a number token as either an integer or float token
        Format: ("-")? & (INT)+ & ("." & (INT)+)?

        Returns:
            Token: integer or float token with TypedToken.INT or TypedToken.FLOAT
        """
        is_float = False

        self.recorder.lock()
        if self.get_char() == "-":
            self.advance()

        while self.get_char().isdigit():
            self.advance()

        # float type, assuming it is not ".." which indicates a range
        # a range will be separated into its individual tokens (number, range delimiter, number)
        if self.get_char() == "." and self.get_chars(2) != "..":
            is_float = True
            self.advance()
            while self.get_char().isdigit():
                self.advance()

        # exponential value, meaning that the type of float will now be TypedToken.EXPONENTIAL_FLOAT
        if self.get_char().lower() == "e":
            is_float = True
            self.advance()

            # allows either + or - after e or E
            if self.get_char().lower() in ("+", "-"):
                self.advance()

            # requires at least one digit
            if not self.get_char().isdigit():
                self.error(f"Got {self.get_char()!r} but expected a digit after an exponential 'e' or 'E' for a float")
            while self.get_char().isdigit():
                self.advance()

        token_type = TypedToken.FLOAT if is_float else TypedToken.INT
        return self.create_new_token(token_type, unlock=True)

    def get_curly_bracket_tag(self, array=False):
        """
        Get all relevant nbt tag or json tokens

        This also checks for proper encapsulation of {} and []
        """
        # gets the beginning bracket token
        if array:
            begin_type = DelimiterToken.OPEN_SQUARE_BRACKET
            end_type = DelimiterToken.CLOSE_SQUARE_BRACKET
        else:
            begin_type = DelimiterToken.OPEN_CURLY_BRACKET
            end_type = DelimiterToken.CLOSE_CURLY_BRACKET

        if not self.current_chars_are(begin_type.value):
            self.error(expected=begin_type.value)
        yield self.create_new_token(begin_type, advance=True)

        while not self.current_chars_are(end_type.value):
            if self.get_char().isspace() and self.get_char() != WhitespaceToken.NEWLINE.value:
                self.skip_whitespace()
                continue

            # skips comments inside nbt tags
            if self.current_chars_are(WhitespaceToken.NEWLINE.value):
                self.advance()
                self.skip_whitespace()
                if self.current_chars_are(WhitespaceToken.COMMENT.value):
                    self.skip_comment()
                continue

            elif self.get_char() in (DelimiterToken.COLON.value, DelimiterToken.COMMA.value, DelimiterToken.SEMICOLON.value):
                yield self.create_new_token(DelimiterToken(self.get_char()), advance=True)

            elif self.current_chars_are(DelimiterToken.OPEN_CURLY_BRACKET.value):
                # recursive call to get nbt in {}
                yield from self.get_curly_bracket_tag()

            elif self.current_chars_are(DelimiterToken.OPEN_SQUARE_BRACKET.value):
                # recursive call to get nbt in []
                yield from self.get_curly_bracket_tag(array=True)

            elif self.get_char() == DelimiterToken.QUOTE.value:
                yield self.get_literal_string(allow_newline=True)
            elif self.get_char().isdigit() or self.get_char() == "-":
                yield self.get_number()
            elif self.get_char().isalpha() or self.get_char() in "_.":
                yield self.get_until_space(exempt_chars=TokenValues.get(DelimiterToken))
            else:
                self.error(f"Invalid character {self.get_char()!r} inside tag")

        # gets the ending bracket token
        yield self.create_new_token(end_type, advance=True)

    def get_block_states(self):
        """
        Gets the minimal amount of tokens required for a block states area
        """
        yield self.create_new_token(DelimiterToken.OPEN_SQUARE_BRACKET, advance=True)

        while not self.current_chars_are(DelimiterToken.CLOSE_SQUARE_BRACKET.value):
            if self.get_char().isspace() and self.get_char() != WhitespaceToken.NEWLINE.value:
                self.skip_whitespace()
                continue

            if self.get_char().isalpha() or self.get_char() == "*":
                yield self.get_until_space(exempt_chars=TokenValues.get(DelimiterToken))
            elif self.get_char().isdigit() or self.get_char() == "-":
                yield self.get_number()
            elif self.get_char() in (DelimiterToken.COMMA.value, DelimiterToken.EQUALS.value):
                yield self.create_new_token(DelimiterToken(self.get_char()), advance=True)
            else:
                self.error(f"Invalid character {self.get_char()!r} inside block states")

        yield self.create_new_token(DelimiterToken.CLOSE_SQUARE_BRACKET, advance=True)

if __name__ == "__main__":
    # import timeit

    with open("log/after_pre_pyexpander.txt") as file:
        text = file.read()

    lexer = Lexer(text)

    for token in lexer:
        logging.debug(repr(token))

    print(repr(lexer))

    # gets only selector tokens
    # lexer = Lexer("@e[x=5,y=4,z=2, type=armor_stand,_entity, _pl=(5..6), {CustomNameVisible:1b}]")
    # for token in lexer.get_selector():
    #     logging.debug(repr(token))

    # lexer = Lexer(r'{"test":1 "test":2}')
    # for token in lexer.get_curly_bracket_tag():
    #     print(repr(token))
    #     logging.debug(repr(token))

    # lexer = Lexer("_pl")
    # print(lexer.get_until_space())
    # lexer = Lexer("@s _ti += players _ti2")
    # lexer = Lexer("setblock _ti <= players _ti2")
    # lexer = Lexer("stonebrick[variant=mossy_stonebrick]")
    # for token in lexer.get_command():
    #     print(repr(token))




if __name__ == "__main__":
    import sys
    sys.path.append("..")
    del sys

    # note that this is actually used since importing it sets up the logging
    # pylint: disable=unused-import
    import fenalib.logging_setup as logging_setup

import os
import json
import logging
from types import MappingProxyType

from fenalib.assert_utils import assert_type

"""
Gets all config data from the config folder
"""

PATH_TO_CONFIG_DIR = "config/"
MAIN_CONFIG_NAME = "config.ini"

class ConfigData:
    """
    Singleton class based off of:
        https://colab.research.google.com/drive/1eajT5Rl9tA-7RmSHMME54B81U5aSKSni#scrollTo=nfQZINGxuUdK

    Sets its own attributes based off of:
        https://stackoverflow.com/questions/2466191/set-attributes-from-dictionary-in-python

    Attributes:
        version (str)
        ego (bool)
        leading_commands (list of strs)
        plugin_conflict_commands (list of strs)

        selector_variable_specifiers (list of strs)
        selector_arguments (list of strs)
        selector_replacements (dict): maps the shorthand version to the minecraft selector argument
        selector_argument_details (dict): maps each selector argument to the parse type and any other details
            - Uses the command json parse type

        execute_dimensions (dict)
        execute_comparison_operators (dict)
        execute_keywords (list)
        execute_data_types (list)

        scoreboard_math (dict)
        scoreboard_special (dict)

        replaceitem_entity_slots (list)
        replaceitem_block_slots (list)

        blocks (list of strs)
        command_names (list of str objects): All possible command names
        bossbar_get (list of str objects)
        bossbar_set (dict)
        effects (list of strs)
        entities (list of strs)
        items (list of str objects)
        team_options (dict): maps each team option to its possible values
    """

    __slots__ = [
        "version",
        "ego",
        "leading_commands",
        "plugin_conflict_commands",
        "selector_variable_specifiers",
        "selector_arguments",
        "selector_replacements",
        "selector_argument_details",
        "execute_dimensions",
        "execute_keywords",
        "execute_data_types",
        "execute_comparison_operators",
        "scoreboard_math",
        "scoreboard_special",
        "replaceitem_entity_slots",
        "replaceitem_block_slots",
        "blocks",
        "bossbar_set",
        "bossbar_get",
        "command_names",
        "effects",
        "entities",
        "items",
        "team_options",
    ]

    def __init__(self, **options):
        if options:
            for key in options:
                setattr(self, key, options[key])

            # makes sure that all variables in __slots__ are used
            for key in ConfigData.__slots__:
                assert hasattr(self, key)

    # options are automatically passed with __new__
    # pylint: disable=unused-argument
    def __new__(cls, **options):
        """
        Ensures they are the same class
        """
        if not hasattr(cls, '_config_data'):
            cls._config_data = super().__new__(cls)
        return cls._config_data

    @property
    def variables(self):
        """
        Returns:
            MappingProxyType: Read-only dictionary of the variables stored in the config data
        """
        variables = {}
        for key in ConfigData.__slots__:
            variables[key] = getattr(self, key)
        return MappingProxyType(variables)
        # variables = {key: getattr(self, key) for key in ConfigData.__slots__)

    def __str__(self):
        # return f"ConfigData[vars={vars(self)}]"
        return f"ConfigData[vars={self.variables}]"

    __repr__ = __str__

def _get_file_str(file_name):
    """
    Gets the data in the file string with the file path relative to this python file (config_data.py)

    Returns:
        str: The full file as a str
    """
    dir_path = os.path.dirname(__file__)
    file_path = os.path.join(dir_path, PATH_TO_CONFIG_DIR + file_name)

    with open(file_path, "r") as file:
        file_data = file.read()

    return file_data


def _get_config_data(version=None):
    """
    Gets all data from config.ini

    For each line:
     - splits with = to get the left hand side (LHS) and right hand side (RHS)
     - splits on "," at the RHS to get all the config options
     - creates the first of the singleton class of ConfigData

    Args:
        version (str or None): The file contents for testing purposes

    Returns:
        dict: All options found in config.ini
    """
    assert_type(version, str, optional=True)

    # the config file should be placed one directory below
    file_data = _get_file_str(MAIN_CONFIG_NAME)
    options = {}

    # gets rid of empty lines, trailing whitespace and comments
    lines = [line.strip() for line in file_data.splitlines() if line.strip() if line.strip()[0] != "#"]

    # gets the list as a split through "," on the RHS unless it is the version
    for line in lines:
        option, data = line.split("=")
        if option == "version":
            options[option] = (data if version is None else version)
        elif option == "ego":
            options[option] = data
        else:
            options[option] = data.split(",")

    return options


def _get_json_config_data(version):
    """
    Gets data from:
        blocks.json
        bossbar.json
        command_names.json
        effects.json
        entities.json
        execute.json
        items.json
        scoreboard.json
        replaceitem.json
        selector.json
        team_options.json

    Returns:
        dict: Maps "blocks", "command_names" and "entities" to their respective lists
    """
    # result for all files
    json_options = {}
    option_names = ("blocks", "bossbar", "command_names", "effects", "entities", "execute", "items", "selector", "scoreboard", "replaceitem", "team_options")

    for option_name in option_names:
        file_name = f"{option_name}.json"
        file_str = _get_file_str(file_name)
        json_object = json.loads(file_str)

        # if it's a string, it refers back to the previous version completely
        # note that this does not prevent recursion
        # use 'temp_version' to not change the actual version variable for the next usage
        temp_version = version
        while isinstance(json_object[temp_version], str):
            temp_version = json_object[temp_version]

        # when config data has new versions
        if isinstance(json_object[temp_version], list):
            json_options[option_name] = json_object[temp_version]
        elif isinstance(json_object[temp_version], dict):
            for option in json_object[temp_version]:
                json_options[option] = json_object[temp_version][option]
        elif json_object[temp_version] is None:
            # doesn't do anything since there is no need to replace None with a different value
            pass
        else:
            raise SyntaxError("Unexpected default case")

    return json_options


def _validate_options(options, valid_options):
    """
    General function to validate that all valid options are found inside the options with no extra options

    Args:
        options (dict): All options found in the config file
        valid_options (frozenset): All expected options
    """
    # checks for a missing option using set difference
    missing_options = valid_options - frozenset(options.keys())
    extra_options = frozenset(options.keys()) - valid_options

    if missing_options:
        raise SyntaxError(f"{set(missing_options)}: Option(s) were not found")

    if extra_options:
        raise SyntaxError(f"{set(extra_options)}: Extra option(s) were detected")


def get_all_data(version=None):
    """
    Main function to get all data from all config files
    If a minecraft version needs to be specified, use this function with the requested version

    Args:
        version (str or None): Optional version to force the config to be read with a specific version
    """
    # gets all config data from config.ini and validates
    general_options = _get_config_data(version=version)
    valid_general_options = frozenset({"version", "ego", "leading_commands", "plugin_conflict_commands"})
    _validate_options(general_options, valid_general_options)

    # makes sure the version is correct
    valid_versions = frozenset({"1.12", "1.13"})
    version = general_options["version"]
    if version not in valid_versions:
        raise SyntaxError(f"{version}: Invalid version in config (must be in {valid_versions})")

    # makes sure "ego" is set to true or false (with any capitalization)
    valid_bool = frozenset({"true", "false"})
    ego = general_options["ego"].lower()
    if ego in valid_bool:
        general_options["ego"] = (ego == "true")
    else:
        raise SyntaxError(f"{ego}: Invalid ego boolean value (must be in {valid_bool})")

    # gets all selector config options from the config data
    # selector_options = _get_selector_config_data(version)

    # gets all blocks, commands and entities
    json_options = _get_json_config_data(version)

    # ensures that this is the first instance of the ConfigData object
    all_options = {**general_options, **json_options}
    config_data = ConfigData(**all_options)
    logging.debug(f"Got the config data: {config_data}")


get_all_data()

if __name__ == "__main__":
    get_all_data(version="1.12")
    print(json.dumps(dict(ConfigData().variables), indent=4))
    print(list(ConfigData().team_options))
    get_all_data(version="1.13")
    print(list(ConfigData().team_options))
    # print(ConfigData().variables)
    # print(list(ConfigData().variables))
    # print(ConfigData().replaceitem_entity_slots)
    # print(config_data)
    # test()





"""
Simply provides the encode/decode function to turn a literal string into a regular string and back
"""

import codecs

def decode_str(string):
    """
    Examples:
        >>> decode_str(r'"hello"')
        'hello'
        >>> decode_str(r'""')
        ''
        >>> decode_str(r'"\\x89\\t"')
        '\\x89\\t'

        >>> decode_str(r'')
        Traceback (most recent call last):
            ...
        SyntaxError: Expected the string '' to be longer than two characters so it can begin and end with a quotation

        >>> decode_str(r'hello')
        Traceback (most recent call last):
            ...
        SyntaxError: Expected the string 'hello' to begin with a quotation

        >>> decode_str(r'"hello')
        Traceback (most recent call last):
            ...
        SyntaxError: Expected the string '"hello' to end with a quotation
    """
    if not len(string) >= 2:
        raise SyntaxError(f"Expected the string {string!r} to be longer than two characters so it can begin and end with a quotation")
    if not string.startswith('"'):
        raise SyntaxError(f"Expected the string {string!r} to begin with a quotation")
    if not string.endswith('"'):
        raise SyntaxError(f"Expected the string {string!r} to end with a quotation")

    return codecs.decode(string[1:-1], "unicode_escape")

def encode_str(string):
    """
    Examples:
        >>> encode_str('hello')
        '"hello"'
        >>> encode_str('\\x89\\t')
        '"\\\\x89\\\\t"'
    """
    return '"' + codecs.encode(string, "unicode_escape").decode("utf-8").replace('"', r'\"') + '"'

if __name__ == "__main__":
    import doctest
    doctest.testmod()

    print(encode_str(r'{CustomName:"{\"text\":\"test\"}"}'))
    print(encode_str(r'{CustomName:"\n"}'))
    print(decode_str(r'"{CustomName:\"\\n\"}"'))
    print(encode_str(decode_str(r'"{CustomName:\"\\n\"}"')))


